/*
 * This is a simple batch-mode recording program.
 *
 * The basic recording functionality was taken from wavnorm 0.4.  This was
 * written by "Duncan" (duncan@zog.net.au).
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <limits.h>
#include <time.h>
#include <sys/time.h>

#include <math.h>
#include <signal.h>

#include "wavutils.h"
#include "text_display.h"
#include "version.h"

/* ======================================================================== */

/** this is for recording to CD so this is the only setting I want **/
#define SAMPLES_PER_SECOND 44100
#define BITS_PER_SAMPLE    16
#define NUMBER_OF_CHANNELS 2

#define BYTES_PER_SAMPLE   (BITS_PER_SAMPLE/8)
#define BYTES_PER_SECOND   (SAMPLES_PER_SECOND*NUMBER_OF_CHANNELS*BYTES_PER_SAMPLE)

#define DEFAULT_DSP_DEVNAME "/dev/dsp"
#define DEFAULT_TIMEOUT     2

/* VU meter timings in seconds.  For best results MAX_HOLDTIME should be a */
/* multiple of VU_REFRESH. */
#define MAX_HOLDTIME        0.20
#define VU_REFRESH          0.05

char *dsp_devname = NULL;
size_t dsp_blocksize = 0;
int append = 0;
int user_exit_req = 0;
int running = 1;
int null_recording = 0;
float threshold,timeout = -1.0;
int restart_freq = -1;
int multiple_recording = 0;
int stdout_output = 0;
int timestamp = 0;
char *filename = NULL;

/* ======================================================================== */

void sig_handler(int signum) {
/*
 * Purpose: handles signals caught by the program:
 *           * SIGINT  = exit program
 *           * SIGALRM = stop the current file and start a new one
 *
 */
  switch (signum) {
    case SIGINT:
      user_exit_req = 1;
      break;
    case SIGALRM:
      running = 0;
      break;
  }
  /* Don't reinstall signal handler for INT signal - this allows emergency */
  /* bailout to proceed if needed. */
  if (signum != SIGINT) {
    signal(signum,sig_handler);
  }
}
/* ======================================================================== */
/* Routines to convert to/from dB.  Assumes incoming data ranges from
 * -32768 to +32767.
 */

float to_dB(float source, int *sign) {
  if (sign != NULL)
    *sign = (source > 0)?1:-1;
  if (source != 0)
    return(20.0 * log10(fabs(source)/32767.0));
  else
    return(20.0 * log10(0.5/32767.0));
}
float from_dB(float source) {
  if (source >= -96)
    return(pow(10,source/20.0)*32767.0);
  else
    return(0);
}
/* ======================================================================== */

int open_dsp_read(char *dsp_filename) {
/*
 * Purpose: opens up /dev/dsp to read audio data.
 *
 * Notes: Presently forces 16 bit stereo at 44100Hz.
 *        Returns the fd to the dsp and sets the block size in blk_size.
 */
int dspfd;
int status;
int arg;

  /** open the sound device and set it up **/
  dspfd = open(dsp_filename,O_RDONLY);
  if (dspfd < 0) {
    fprintf(stderr,"Error: open(%s) %s\n", dsp_filename,strerror(errno));
    return(-1);
  }
  arg = BITS_PER_SAMPLE;
  status = ioctl(dspfd, SOUND_PCM_WRITE_BITS, &arg);

  if (status == -1) {
    perror("SOUND_PCM_WRITE_BITS ioctl failed");
    return(-1);
  }
  if (arg != BITS_PER_SAMPLE) {
    perror("unable to set sample size");
    return(-1);
  }

  arg = NUMBER_OF_CHANNELS;  /* mono or stereo */
  status = ioctl(dspfd, SOUND_PCM_WRITE_CHANNELS, &arg);
  if (status == -1) {
    perror("SOUND_PCM_WRITE_CHANNELS ioctl failed");
    return(-1);
  }
  if (arg != NUMBER_OF_CHANNELS) {
    perror("unable to set number of channels");
    return(-1);
  }

  arg = SAMPLES_PER_SECOND;      /* sampling rate */
  status = ioctl(dspfd, SOUND_PCM_WRITE_RATE, &arg);
  if (status == -1) {
    perror("SOUND_PCM_WRITE_RATE ioctl failed");
    return(-1);
  }
  status = ioctl(dspfd, SNDCTL_DSP_GETBLKSIZE,&arg);
  /* fprintf(stderr,"status %d, blksize %d\n",status,arg); */
  dsp_blocksize = arg;

  /* 100ms delay to avoid the blips when opening the dsp device */
  usleep(100*1000);

  fprintf(stderr,"Input device is %s\n",dsp_filename);
  fprintf(stderr,"Sampling %d channel(s)\n",NUMBER_OF_CHANNELS);
  fprintf(stderr,"%d Hz sampling frequency, %d bits per sample\n",
    SAMPLES_PER_SECOND,BITS_PER_SAMPLE);
  fprintf(stderr,"Using block size of %d bytes\n",dsp_blocksize);
  if (timeout==-1 && restart_freq<=0) 
    fprintf(stderr,"Recording until <ctrl>-C is pressed\n");
  if (restart_freq>0) {
    fprintf(stderr,"Restarting every %d seconds\n",restart_freq);
  }
  fprintf(stderr,"\n");
  return(dspfd);
}
/* ======================================================================== */

void make_length_str(int n_samples, char *str) {
/*
 * Purpose: constructs a string describing the length of the current sample.
 *          "str" must point to a string buffer with space for at least 20 
 *          characters.
 *
 * Notes: assumes stereo recording with a sampling frequency of 44100Hz.
 */
int sec100s = (int)(n_samples / (SAMPLES_PER_SECOND/100.0));
int size = NUMBER_OF_CHANNELS * n_samples * BYTES_PER_SAMPLE;
char sizebuf[10];

  if (size < 9999)
    sprintf(sizebuf,"%4dB",size);
  else
  if (size < 9999*1024)
    sprintf(sizebuf,"%4dkB",size/1024);
  else
  if (size < 99.9*1024*1024)
    sprintf(sizebuf,"%4.1fMB",size/(1024*1024.0));
  else
  if (size < 2047*1024*1024)
    sprintf(sizebuf,"%4dMB",size/(1024*1024));
  else
    sprintf(sizebuf,"!!!!MB");

  sprintf(str,"%02d:%02d.%02d (%s)",sec100s/(100*60),(sec100s/100)%60,
    sec100s % 100,sizebuf);
}
/* ======================================================================== */

int record_wav_file(int dspfd, char *filename) {
/*
 * Purpose: records into a single given file.
 *
 */
char *wavbuf;
int wavfd;
size_t curr_blocksize;
int recording = (timeout==-1);
float below_age = 0;
size_t bytes_written = 0;
char length_str[20];
         
WAVE_HEADER wav_header, *wav_header_on_disk;
short max[2], ave[2];
int acc_max[2], acc_ave[2];
int vu_count, i;
int max_age[2];

  /* If the a start time quantisation is desired, set the interval timer */
  /* to go off at the time which the *next* file should start (ie: when  */
  /* this one should end). */
  if (restart_freq > 0) {
    struct timeval tv;
    struct timezone tz;
    struct itimerval itv;
    int local_time;
    gettimeofday(&tv,&tz);
    local_time = tv.tv_sec - (tz.tz_minuteswest*60);

    /* Calculate the time (in seconds) to the start of the *next* file */
    itv.it_value.tv_sec = restart_freq - (local_time % restart_freq);
    itv.it_value.tv_usec = 0;
    if (itv.it_value.tv_sec <= 1)
      itv.it_value.tv_sec += restart_freq;
    /* Set the timer interval, which is just restart_freq */
    itv.it_interval.tv_sec = restart_freq;
    itv.it_interval.tv_usec = 0;

    /* Set up the interval timer */
    setitimer(ITIMER_REAL,&itv,NULL);
  }

  /* Reset running status */
  running = 1;

  if ((wavbuf = malloc(dsp_blocksize)) == NULL) {
    fprintf(stderr,"Error: malloc() %s\n",strerror(errno));
    return(0);
  }

  if (stdout_output) {
    wavfd = fileno(stdout);
  } else {
    create_wav_header(&wav_header,
      NUMBER_OF_CHANNELS,SAMPLES_PER_SECOND,BITS_PER_SAMPLE);
    if (!append) {
      wavfd = open(filename,O_RDWR | O_CREAT | O_TRUNC,0600);
    } else {
      wavfd = open(filename,O_RDWR | O_CREAT, 0600);
    }
  }
  if (wavfd == -1) {
    fprintf(stderr,"Error: open() %s\n",strerror(errno));
    free(wavbuf);
    return(0);
  }
  if (!stdout_output) {
    /* Go to end of file.  If position isn't 0 it means we're appending and */
    /* there's no need to write the file header. */
    i = lseek(wavfd,0,SEEK_END);
    if (i<0) {
      fprintf(stderr,"Error: %s\n",strerror(errno));
      close(wavfd);
      free(wavbuf);
      return(0);
    }
    if (i==0) {
      if (write(wavfd,(void *)&wav_header,sizeof(wav_header)) != sizeof(wav_header)) {
        fprintf(stderr,"Error: write() %s\n",strerror(errno));
        close(wavfd);
        free(wavbuf);
        return(0);
      }
    } else {
      /* Append mode, so set bytes_written to reflect bytes already present */
      bytes_written = i - sizeof(wav_header);
    }
  }
  if (timeout > 0) {
    fprintf(stderr,"Ready to record %s\n\n",filename);
  }
  text_vu_headers(); 
  vu_count = 0;
  for (i=0;i<2;i++) {
    max_age[i] = 0;
    acc_ave[i] = acc_max[i] = 0;
  }
  while (running) {
    /* A short read could be due to a signal interrupt so keep track of */
    /* the block size actually read */
    curr_blocksize = read(dspfd,wavbuf,dsp_blocksize);
    if (curr_blocksize!=dsp_blocksize && errno!=0) {
      fprintf(stderr,"read(dsp) %s",strerror(errno));
      return(0);
    }
    get_stereo_peak_values((short *)wavbuf,(short *)(wavbuf + curr_blocksize),
      &max[0], &max[1]);
    get_stereo_ave_values((short *)wavbuf,(short *)(wavbuf + curr_blocksize),
      &ave[0], &ave[1]);
    if (timeout>0) {
      recording = (recording || to_dB(max[0],NULL)>threshold ||
        to_dB(max[1],NULL)>threshold);
      if (recording) {
        if (to_dB(max[0],NULL)<threshold && to_dB(max[1],NULL)<threshold)
          below_age += curr_blocksize;
        else
          below_age = 0;
        running = (below_age/BYTES_PER_SECOND < timeout);
      }
    }
    if (!null_recording && recording) {
      if (write(wavfd,wavbuf,curr_blocksize) != curr_blocksize) {
        fprintf(stderr,"write(%s) %s",filename,strerror(errno));
        return(0);
      }
      bytes_written += curr_blocksize;
    }

    for (i=0;i<2;i++) {
      if (max[i]>acc_max[i] || (float)max_age[i]/BYTES_PER_SECOND > MAX_HOLDTIME) {
        acc_max[i] = max[i];
        max_age[i] = 0;
      } else {
        max_age[i] += curr_blocksize;
      }
      acc_ave[i] += ave[i];
    }
    vu_count++;
    if ((float)vu_count*dsp_blocksize/BYTES_PER_SECOND > VU_REFRESH) {
      text_vu_values(to_dB(acc_max[0],NULL),to_dB(acc_max[1],NULL),
                     to_dB((float)acc_ave[0]/vu_count,NULL),
                     to_dB((float)acc_ave[1]/vu_count,NULL));
      vu_count = 0;
      for (i=0;i<2;i++)
        acc_ave[i] = 0;

      if (recording) {
        make_length_str(bytes_written/(BYTES_PER_SAMPLE*NUMBER_OF_CHANNELS),
          length_str);
        fprintf(stderr,length_str);
      } else
        fprintf(stderr,"waiting...");
      fflush(stderr);
    }
    if (user_exit_req) {
      running = 0;
    }
  }
  fprintf(stderr,"\n\n");
  if (bytes_written!=0) {
    make_length_str(bytes_written/(BYTES_PER_SAMPLE*NUMBER_OF_CHANNELS),
      length_str);
    if (stdout_output) {
      fprintf(stderr,"Recording to stdout complete: final length %s\n",length_str);
    } else {
      fprintf(stderr,"Recording of %s complete: final length %s\n",filename,length_str);
    }
  }

  /* Set the wav file header to have the right length */
  if (!stdout_output) {
    if (bytes_written == 0) {
      fprintf(stderr,"Recording of %s finished with no data written: file erased.\n\n",filename);
      remove(filename);
    } else {
      wav_header_on_disk = (WAVE_HEADER *)mmap(NULL,sizeof(WAVE_HEADER),
        PROT_READ | PROT_WRITE,MAP_SHARED,wavfd,0);
      if (wav_header_on_disk == MAP_FAILED) {
        fprintf(stderr,"Warning: mmap() %s\n",strerror(errno));
        fprintf(stderr,"File %s will be OK but you may have to fix the length\n",filename);
      } else {
        wav_header_on_disk->nDataBytes = bytes_written;
        /* RiffSize = the size of all blocks in wav file including their */
        /* headers.  There's only one block in these files, so only one  */
        /* 36 byte header.  WAV header is 44 bytes: 8 bytes for the RIFF */
        /* header, 36 for the 1st (and only) block header. */
        wav_header_on_disk->RiffSize = bytes_written + 36;
        munmap(wav_header_on_disk,sizeof(WAVE_HEADER));
      }
    }
  }
  if (!stdout_output)
    close(wavfd);               

  free(wavbuf);
  return(1);
}
/* ======================================================================== */

void show_usage() {
/*
 * Purpose: gives the user some help on the program's functionality.
 *
 */

  fprintf(stderr,"batchrec version " VERSION "\n");
  fprintf(stderr,"(c) 2001-2002 Jonathan Woithe (jwoithe@physics.adelaide.edu.au)\n"
                 "Released under the terms of the GNU General Public License.  No warranty is\n"
                 "given or implied, not even for fitness of purpose; see source for full details.\n\n"
                 "usage: batchrec [options] <filename>\n"
                 "Available options:\n"
                 " -a                  If output file exists, append rather than truncating.\n"
                 "                     Playback of output file will contain strange effects if\n"
                 "                     existing data format and rate doesn't match batchrec's.\n"
                 " -d device           Use <device> as audio source\n"
                 " -h                  Display this help and exit\n"
                 " -n                  Null recording: do everything except write data to file\n"
                 " -r restart_freq     Multiple recording mode: start new recording every\n"
                 "                     <restart_freq> seconds\n"
                 " -T                  Timestamp the output filename\n"
                 " -t thres[,timeout]  Set threshold and timeout for multiple recording mode\n"
                 "\n\n");
  fprintf(stderr,"In normal mode, recording is made into the supplied filename until <ctrl>-C\n"
                 "is pressed.  If <filename> is \"-\", raw interleaved 16 bit PCM data is\n"
                 "written to stdout.\n\n");
  fprintf(stderr,"In multiple recording mode, recording is made to a series of files.  A new\n"
                 "file is started for one of several reasons:\n"
                 "  * if `-t' is specified, recording starts when the input exceeds the given\n"
                 "    threshold (in dB) and continues until the signal stays below the threshold\n"
                 "    for a period of time equal to the timeout (default is %d seconds).\n"
                 "  * if `-r' is specified a restart if forced every <restart_freq> seconds\n"
                 "    aligned with the local time of day: if <restart_freq> is 3600 for example,\n"
                 "    a new file is started every hour on the hour.\n"
                 "If both `-t' and `-r' is specified, the `-r' operation will start a new file\n"
                 "even if the threshold rule is not applicable.\n"
                 "Multiple recording mode continues until <ctrl>-C is pressed.  In this mode,\n"
                 "the output files are generated by appending the relevant recording number to\n"
                 "the supplied filename.  If the supplied filename ends in a valid number, that\n"
                 "number (zero padded to 4 digits) is used as the initial recording number.  The\n"
                 "default timeout value is 2 seconds.\n\n"
                 "Currently recording is fixed at 44.1kHz/16 bit format.\n",DEFAULT_TIMEOUT);
}
/* ======================================================================== */

int parse_args(int argc, char *argv[]) {
/*
 * Purpose: parses the command line parameters.  Returns 1 on success.
 *
 */
int c;
int error = 0;
  while (!error && (c = getopt(argc, argv, "ad:hnr:Tt:"))!=EOF) {
    switch(c) {
      case 'a':
        append = 1;
        break;
      case 'd':
        dsp_devname = strdup(optarg);
        break;
      case 'h':
        show_usage();
        return(0);
        break;
      case 'n':
        null_recording = 1;
        break;
      case 'r': {
        int freq;
        if (sscanf(optarg,"%d",&freq)!=1 || freq<=0) {
          error = 1;
          fprintf(stderr,"%s: malformed `-r' parameter\n",argv[0]);
        } else {
          restart_freq = freq;
        }
        break;
      }
      case 'T': {
        timestamp = 1;
        break;
      }
      case 't': {
        char sep;
        int num = sscanf(optarg,"%g%c%g",&threshold,&sep,&timeout);
        if (num == 1)
          timeout = DEFAULT_TIMEOUT;
        else
        if (num==2 || sep!=',') {
          error = 1;
          fprintf(stderr,"%s: malformed `-t' parameter\n",argv[0]);
        } else
        if (threshold>0 || timeout<=0) {
          error = 1;
          fprintf(stderr,"%s: `-t' parameter values invalid\n",argv[0]);
        }
        break;
      }
      default:
        error = 1;
    }
  }
  if (!error && dsp_devname==NULL)
    dsp_devname = strdup(DEFAULT_DSP_DEVNAME);

  if (!error && (optind + 1)==argc) {
    filename = strdup(argv[optind]);
  } else {
    error = 1;
    fprintf(stderr,"%s: no output filename specified\n",argv[0]);
  }
  stdout_output = (filename!=NULL) && !strcmp(filename,"-");
  multiple_recording = (timeout>0 || restart_freq>0);
  if (stdout_output) {
    if (multiple_recording) {
      error = 1;
      fprintf(stderr,"%s: multiple recordings makes no sense with output to stdout\n",argv[0]);
    }
    if (append) {
      error = 1;
      fprintf(stderr,"%s: appending makes no sense with output to stdout\n",argv[0]);
    }
    if (timestamp) {
      error = 1;
      fprintf(stderr,"%s: timestamping makes no sense with output to stdout\n",argv[0]);
    }
  }
  return(!error);
}
/* ======================================================================== */

void cleanup() {
/*
 * Purpose: Performs final cleanup duties prior to exit.
 *
 */
  if (dsp_devname != NULL)
    free(dsp_devname);
  if (filename != NULL)
    free(filename);
}
/* ======================================================================== */

char *next_filename(char *name,int next) {
/*
 * Purpose: generates a sequential filename for use in multiple recording
 *          mode.  If "next" is true, the new filename will be the next number
 *          along (if the filename contains a number), or 0 if no number is
 *          present.  If "next" is false, the filename provided is simply
 *          made compatible with the sequential naming convention.
 *
 *          In timestamp mode the "number" is generated from the current
 *          time.
 */
char *baseend,*ext;
char *res;
int i = 0;
int num = 0;
int len;

  if ((baseend = strchr(name,'.')) == NULL) {
    ext = baseend = name + strlen(name);
  } else {
    ext = baseend+1;
    *baseend = '\0';
  }
  while (i<(timestamp?14:4) && *(baseend-1)>='0' && *(baseend-1)<='9') {
    i++;
    baseend--;
  }
  /* A number exists, so identify it */
  if (i>0) {
    sscanf(baseend,"%d",&num);
    if (next)
      num++;
    *baseend = '\0';
  }
  if (!timestamp)
    len = strlen(name)+4+1+strlen(ext)+1;
  else
    len = strlen(name)+14+1+strlen(ext)+1;
  res = malloc(len);
  if (!timestamp)
    snprintf(res,len,"%s%04d%s%s",name,num,strlen(ext)>0?".":"",ext);
  else {
    time_t t = time(NULL);
    struct tm *tm;
    tm = localtime(&t);
    snprintf(res,len,"%s%04d%02d%02d%02d%02d%02d%s%s",name,
      tm->tm_year+1900,tm->tm_mon+1,tm->tm_mday,
      tm->tm_hour,tm->tm_min,tm->tm_sec,
      strlen(ext)>0?".":"",ext);
  }
  res[len-1]=0;
  free(name);
  return(res);
}
/* ======================================================================== */

int main(int argc,char *argv[]) {
int dspfd,c;

  /* Trap selected signals */
  signal(SIGINT,sig_handler);
  signal(SIGALRM,sig_handler);

  if ((c = parse_args(argc,argv))) {
    /* Open the dsp device */
    dspfd = open_dsp_read(dsp_devname);
    if (dspfd == -1) {
      fprintf(stderr,"Error getting %s ready for recording\n",dsp_devname);
    } else {
      if (multiple_recording || timestamp)
        filename = next_filename(filename,0);
      while (!user_exit_req) {
        record_wav_file(dspfd,filename);
        if (multiple_recording)
          filename = next_filename(filename,1);
      }
      close(dspfd);
    }
  }
  cleanup();
  return(!c);
}
